"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.update = exports.destroy = exports.create = void 0;
const sequelize_1 = require("sequelize");
const actions_1 = require("@nocobase/actions");
function create(context, next) {
    return __awaiter(this, void 0, void 0, function* () {
        const { db } = context;
        const repository = actions_1.utils.getRepositoryFromParams(context);
        const { whitelist, blacklist, updateAssociationValues, values, associatedIndex: workflowId } = context.action.params;
        context.body = yield db.sequelize.transaction((transaction) => __awaiter(this, void 0, void 0, function* () {
            const workflow = yield repository.getSourceModel(transaction);
            if (workflow.executed) {
                context.throw(400, 'Node could not be created in executed workflow');
            }
            const instance = yield repository.create({
                values,
                whitelist,
                blacklist,
                updateAssociationValues,
                context,
                transaction
            });
            if (!instance.upstreamId) {
                const previousHead = yield repository.findOne({
                    filter: {
                        id: {
                            $ne: instance.id
                        },
                        upstreamId: null
                    },
                    transaction
                });
                if (previousHead) {
                    yield previousHead.setUpstream(instance, { transaction });
                    yield instance.setDownstream(previousHead, { transaction });
                    instance.set('downstream', previousHead);
                }
                return instance;
            }
            const upstream = yield instance.getUpstream({ transaction });
            if (instance.branchIndex == null) {
                const downstream = yield upstream.getDownstream({ transaction });
                if (downstream) {
                    yield downstream.setUpstream(instance, { transaction });
                    yield instance.setDownstream(downstream, { transaction });
                    instance.set('downstream', downstream);
                }
                yield upstream.update({
                    downstreamId: instance.id
                }, { transaction });
                upstream.set('downstream', instance);
            }
            else {
                const [downstream] = yield upstream.getBranches({
                    where: {
                        id: {
                            [sequelize_1.Op.ne]: instance.id
                        },
                        branchIndex: instance.branchIndex
                    },
                    transaction
                });
                if (downstream) {
                    yield downstream.update({
                        upstreamId: instance.id,
                        branchIndex: null
                    }, { transaction });
                    yield instance.setDownstream(downstream, { transaction });
                    instance.set('downstream', downstream);
                }
            }
            instance.set('upstream', upstream);
        }));
        yield next();
    });
}
exports.create = create;
function searchBranchNodes(nodes, from) {
    const branchHeads = nodes
        .filter((item) => item.upstreamId === from.id && item.branchIndex != null);
    return branchHeads.reduce((flatten, head) => flatten.concat(searchBranchDownstreams(nodes, head)), []);
}
function searchBranchDownstreams(nodes, from) {
    let result = [];
    for (let search = from; search; search = search.downstream) {
        result = [...result, search, ...searchBranchNodes(nodes, search)];
    }
    return result;
}
function destroy(context, next) {
    return __awaiter(this, void 0, void 0, function* () {
        const { db } = context;
        const repository = actions_1.utils.getRepositoryFromParams(context);
        const { filterByTk } = context.action.params;
        context.body = yield db.sequelize.transaction((transaction) => __awaiter(this, void 0, void 0, function* () {
            const workflow = yield repository.getSourceModel(transaction);
            if (workflow.executed) {
                context.throw(400, 'Nodes in executed workflow could not be deleted');
            }
            const fields = ['id', 'upstreamId', 'downstreamId', 'branchIndex'];
            const instance = yield repository.findOne({
                filterByTk,
                fields: [...fields, 'workflowId'],
                appends: ['upstream', 'downstream'],
                transaction
            });
            const { upstream, downstream } = instance.get();
            if (upstream && upstream.downstreamId === instance.id) {
                yield upstream.update({
                    downstreamId: instance.downstreamId
                }, { transaction });
            }
            if (downstream) {
                yield downstream.update({
                    upstreamId: instance.upstreamId,
                    branchIndex: instance.branchIndex
                }, { transaction });
            }
            const nodes = yield repository.find({
                filter: {
                    workflowId: instance.workflowId
                },
                fields,
                transaction
            });
            const nodesMap = new Map();
            // make map
            nodes.forEach(item => {
                nodesMap.set(item.id, item);
            });
            // overwrite
            nodesMap.set(instance.id, instance);
            // make linked list
            nodes.forEach(item => {
                if (item.upstreamId) {
                    item.upstream = nodesMap.get(item.upstreamId);
                }
                if (item.downstreamId) {
                    item.downstream = nodesMap.get(item.downstreamId);
                }
            });
            const branchNodes = searchBranchNodes(nodes, instance);
            yield repository.destroy({
                filterByTk: [instance.id, ...branchNodes.map(item => item.id)],
                transaction
            });
            return instance;
        }));
        yield next();
    });
}
exports.destroy = destroy;
function update(context, next) {
    return __awaiter(this, void 0, void 0, function* () {
        const { db } = context;
        const repository = actions_1.utils.getRepositoryFromParams(context);
        const { filterByTk, values, whitelist, blacklist, filter, updateAssociationValues } = context.action.params;
        context.body = yield db.sequelize.transaction((transaction) => __awaiter(this, void 0, void 0, function* () {
            // TODO(optimize): duplicated instance query
            const { workflow } = yield repository.findOne({
                filterByTk,
                appends: ['workflow.executed'],
                transaction
            });
            if (workflow.executed) {
                context.throw(400, 'Nodes in executed workflow could not be reconfigured');
            }
            return repository.update({
                filterByTk,
                values,
                whitelist,
                blacklist,
                filter,
                updateAssociationValues,
                context,
                transaction
            });
        }));
        yield next();
    });
}
exports.update = update;
//# sourceMappingURL=nodes.js.map